@@ -13,12 +13,14 @@
 #include "src/codegen/machine-type.h"
 #include "src/codegen/register-configuration.h"
 #include "src/codegen/source-position.h"
+#include "src/compiler/backend/instruction-codes.h"
 #include "src/compiler/common-operator.h"
 #include "src/compiler/frame-states.h"
 #include "src/compiler/graph.h"
 #include "src/compiler/node.h"
 #include "src/compiler/schedule.h"
 #include "src/compiler/turboshaft/graph.h"
+#include "src/compiler/turboshaft/loop-finder.h"
 #include "src/compiler/turboshaft/operations.h"
 #include "src/deoptimizer/deoptimizer.h"
 #include "src/execution/frames.h"
@@ -79,6 +81,8 @@ FlagsCondition CommuteFlagsCondition(FlagsCondition condition) {
     case kNotOverflow:
     case kUnorderedEqual:
     case kUnorderedNotEqual:
+    case kIsNaN:
+    case kIsNotNaN:
       return condition;
   }
   UNREACHABLE();
@@ -257,6 +261,9 @@ std::ostream& operator<<(std::ostream& os, const InstructionOperand& op) {
         case MachineRepresentation::kWord64:
           os << "|w64";
           break;
+        case MachineRepresentation::kFloat16:
+          os << "|f16";
+          break;
         case MachineRepresentation::kFloat32:
           os << "|f32";
           break;
@@ -284,6 +291,9 @@ std::ostream& operator<<(std::ostream& os, const InstructionOperand& op) {
         case MachineRepresentation::kCompressed:
           os << "|c";
           break;
+        case MachineRepresentation::kProtectedPointer:
+          os << "|pp";
+          break;
         case MachineRepresentation::kIndirectPointer:
           os << "|ip";
           break;
@@ -477,6 +487,10 @@ std::ostream& operator<<(std::ostream& os, const FlagsMode& fm) {
       return os << "trap";
     case kFlags_select:
       return os << "select";
+    case kFlags_conditional_set:
+      return os << "conditional set";
+    case kFlags_conditional_branch:
+      return os << "conditional branch";
   }
   UNREACHABLE();
 }
@@ -531,6 +545,10 @@ std::ostream& operator<<(std::ostream& os, const FlagsCondition& fc) {
       return os << "positive or zero";
     case kNegative:
       return os << "negative";
+    case kIsNaN:
+      return os << "is nan";
+    case kIsNotNaN:
+      return os << "is not nan";
   }
   UNREACHABLE();
 }
@@ -611,7 +629,7 @@ std::ostream& operator<<(std::ostream& os, const Constant& constant) {
     case Constant::kFloat64:
       return os << constant.ToFloat64().value();
     case Constant::kExternalReference:
-      return os << constant.ToExternalReference().address();
+      return os << constant.ToExternalReference();
     case Constant::kHeapObject:  // Fall through.
     case Constant::kCompressedHeapObject:
       return os << Brief(*constant.ToHeapObject());
@@ -716,11 +734,9 @@ static InstructionBlock* InstructionBlockFor(Zone* zone,
   return instr_block;
 }
 
-static InstructionBlock* InstructionBlockFor(Zone* zone,
-                                             const turboshaft::Graph& graph,
-                                             const turboshaft::Block* block) {
-  // TODO(nicohartmann@): Properly get the loop_header.
-  turboshaft::Block* loop_header = nullptr;  // block->loop_header()
+static InstructionBlock* InstructionBlockFor(
+    Zone* zone, const turboshaft::Graph& graph, const turboshaft::Block* block,
+    const turboshaft::Block* loop_header) {
   bool is_handler =
       block->FirstOperation(graph).Is<turboshaft::CatchBlockBeginOp>();
   bool deferred = block->get_custom_data(
@@ -728,7 +744,7 @@ static InstructionBlock* InstructionBlockFor(Zone* zone,
   InstructionBlock* instr_block = zone->New<InstructionBlock>(
       zone, GetRpo(block), GetRpo(loop_header), GetLoopEndRpo(block),
       GetRpo(block->GetDominator()), deferred, is_handler);
-  if (block->HasExactlyNPredecessors(1)) {
+  if (block->PredecessorCount() == 1) {
     const turboshaft::Block* predecessor = block->LastPredecessor();
     if (V8_UNLIKELY(
             predecessor->LastOperation(graph).Is<turboshaft::SwitchOp>())) {
@@ -823,10 +839,16 @@ InstructionBlocks* InstructionSequence::InstructionBlocksFor(
   new (blocks)
       InstructionBlocks(static_cast<int>(graph.block_count()), nullptr, zone);
   size_t rpo_number = 0;
+  // TODO(dmercadier): currently, the LoopFinder is just used to compute loop
+  // headers. Since it's somewhat expensive to compute this, we should also use
+  // the LoopFinder to compute the special RPO (we would only need to run the
+  // LoopFinder once to compute both the special RPO and the loop headers).
+  turboshaft::LoopFinder loop_finder(zone, &graph);
   for (const turboshaft::Block& block : graph.blocks()) {
     DCHECK(!(*blocks)[rpo_number]);
     DCHECK_EQ(RpoNumber::FromInt(block.index().id()).ToSize(), rpo_number);
-    (*blocks)[rpo_number] = InstructionBlockFor(zone, graph, &block);
+    (*blocks)[rpo_number] = InstructionBlockFor(
+        zone, graph, &block, loop_finder.GetLoopHeader(&block));
     ++rpo_number;
   }
   return blocks;
@@ -952,7 +974,10 @@ InstructionSequence::InstructionSequence(Isolate* isolate,
       zone_(instruction_zone),
       instruction_blocks_(instruction_blocks),
       ao_blocks_(nullptr),
-      source_positions_(zone()),
+      // Pre-allocate the hash map of source positions based on the block count.
+      // (The actual number of instructions is only known after instruction
+      // selection, but should at least correlate with the block count.)
+      source_positions_(zone(), instruction_blocks->size() * 2),
       // Avoid collisions for functions with 256 or less constant vregs.
       constants_(zone(), 256),
       immediates_(zone()),
@@ -1015,6 +1040,8 @@ static MachineRepresentation FilterRepresentation(MachineRepresentation rep) {
     case MachineRepresentation::kWord8:
     case MachineRepresentation::kWord16:
       return InstructionSequence::DefaultRepresentation();
+    case MachineRepresentation::kFloat16:
+      return MachineRepresentation::kFloat32;
     case MachineRepresentation::kWord32:
     case MachineRepresentation::kWord64:
     case MachineRepresentation::kTaggedSigned:
@@ -1026,15 +1053,14 @@ static MachineRepresentation FilterRepresentation(MachineRepresentation rep) {
     case MachineRepresentation::kSimd256:
     case MachineRepresentation::kCompressedPointer:
     case MachineRepresentation::kCompressed:
+    case MachineRepresentation::kProtectedPointer:
     case MachineRepresentation::kSandboxedPointer:
       return rep;
     case MachineRepresentation::kNone:
     case MachineRepresentation::kMapWord:
     case MachineRepresentation::kIndirectPointer:
-      break;
+      UNREACHABLE();
   }
-
-  UNREACHABLE();
 }
 
 MachineRepresentation InstructionSequence::GetRepresentation(
@@ -1128,7 +1154,8 @@ namespace {
 size_t GetConservativeFrameSizeInBytes(FrameStateType type,
                                        size_t parameters_count,
                                        size_t locals_count,
-                                       BytecodeOffset bailout_id) {
+                                       BytecodeOffset bailout_id,
+                                       uint32_t wasm_liftoff_frame_size) {
   switch (type) {
     case FrameStateType::kUnoptimizedFunction: {
       auto info = UnoptimizedFrameInfo::Conservative(
@@ -1168,6 +1195,10 @@ size_t GetConservativeFrameSizeInBytes(FrameStateType type,
           config);
       return info.frame_size_in_bytes();
     }
+#if V8_ENABLE_WEBASSEMBLY
+    case FrameStateType::kLiftoffFunction:
+      return wasm_liftoff_frame_size;
+#endif  // V8_ENABLE_WEBASSEMBLY
   }
   UNREACHABLE();
 }
@@ -1176,13 +1207,14 @@ size_t GetTotalConservativeFrameSizeInBytes(FrameStateType type,
                                             size_t parameters_count,
                                             size_t locals_count,
                                             BytecodeOffset bailout_id,
+                                            uint32_t wasm_liftoff_frame_size,
                                             FrameStateDescriptor* outer_state) {
   size_t outer_total_conservative_frame_size_in_bytes =
       (outer_state == nullptr)
           ? 0
           : outer_state->total_conservative_frame_size_in_bytes();
   return GetConservativeFrameSizeInBytes(type, parameters_count, locals_count,
-                                         bailout_id) +
+                                         bailout_id, wasm_liftoff_frame_size) +
          outer_total_conservative_frame_size_in_bytes;
 }
 
@@ -1190,22 +1222,26 @@ size_t GetTotalConservativeFrameSizeInBytes(FrameStateType type,
 
 FrameStateDescriptor::FrameStateDescriptor(
     Zone* zone, FrameStateType type, BytecodeOffset bailout_id,
-    OutputFrameStateCombine state_combine, size_t parameters_count,
-    size_t locals_count, size_t stack_count,
+    OutputFrameStateCombine state_combine, uint16_t parameters_count,
+    uint16_t max_arguments, size_t locals_count, size_t stack_count,
     MaybeHandle<SharedFunctionInfo> shared_info,
-    FrameStateDescriptor* outer_state)
+    FrameStateDescriptor* outer_state, uint32_t wasm_liftoff_frame_size,
+    uint32_t wasm_function_index)
     : type_(type),
       bailout_id_(bailout_id),
       frame_state_combine_(state_combine),
       parameters_count_(parameters_count),
+      max_arguments_(max_arguments),
       locals_count_(locals_count),
       stack_count_(stack_count),
       total_conservative_frame_size_in_bytes_(
           GetTotalConservativeFrameSizeInBytes(
-              type, parameters_count, locals_count, bailout_id, outer_state)),
+              type, parameters_count, locals_count, bailout_id,
+              wasm_liftoff_frame_size, outer_state)),
       values_(zone),
       shared_info_(shared_info),
-      outer_state_(outer_state) {}
+      outer_state_(outer_state),
+      wasm_function_index_(wasm_function_index) {}
 
 size_t FrameStateDescriptor::GetHeight() const {
   switch (type()) {
@@ -1230,6 +1266,10 @@ size_t FrameStateDescriptor::GetHeight() const {
       //   CreateJavaScriptBuiltinContinuationFrameState), and
       // - does *not* include the context.
       return parameters_count();
+#if V8_ENABLE_WEBASSEMBLY
+    case FrameStateType::kLiftoffFunction:
+      return locals_count() + parameters_count();
+#endif
   }
   UNREACHABLE();
 }
@@ -1271,12 +1311,12 @@ size_t FrameStateDescriptor::GetJSFrameCount() const {
 #if V8_ENABLE_WEBASSEMBLY
 JSToWasmFrameStateDescriptor::JSToWasmFrameStateDescriptor(
     Zone* zone, FrameStateType type, BytecodeOffset bailout_id,
-    OutputFrameStateCombine state_combine, size_t parameters_count,
+    OutputFrameStateCombine state_combine, uint16_t parameters_count,
     size_t locals_count, size_t stack_count,
     MaybeHandle<SharedFunctionInfo> shared_info,
     FrameStateDescriptor* outer_state, const wasm::FunctionSig* wasm_signature)
     : FrameStateDescriptor(zone, type, bailout_id, state_combine,
-                           parameters_count, locals_count, stack_count,
+                           parameters_count, 0, locals_count, stack_count,
                            shared_info, outer_state),
       return_kind_(wasm::WasmReturnTypeFromSignature(wasm_signature)) {}
 #endif  // V8_ENABLE_WEBASSEMBLY
@@ -1308,6 +1348,8 @@ std::ostream& operator<<(std::ostream& os, StateValueKind kind) {
       return os << "ArgumentsElements";
     case StateValueKind::kArgumentsLength:
       return os << "ArgumentsLength";
+    case StateValueKind::kRestLength:
+      return os << "RestLength";
     case StateValueKind::kPlain:
       return os << "Plain";
     case StateValueKind::kOptimizedOut:
