@@ -16,16 +16,16 @@
 namespace v8 {
 namespace internal {
 
-void IC::update_lookup_start_object_map(Handle<Object> object) {
+void IC::update_lookup_start_object_map(DirectHandle<Object> object) {
   if (IsSmi(*object)) {
     lookup_start_object_map_ = isolate_->factory()->heap_number_map();
   } else {
     lookup_start_object_map_ =
-        handle(HeapObject::cast(*object)->map(), isolate_);
+        handle(Cast<HeapObject>(*object)->map(), isolate_);
   }
 }
 
-bool IC::IsHandler(MaybeObject object) {
+bool IC::IsHandler(Tagged<MaybeObject> object) {
   Tagged<HeapObject> heap_object;
   return (IsSmi(object) && (object.ptr() != kNullAddress)) ||
          (object.GetHeapObjectIfWeak(&heap_object) &&
