@@ -11,9 +11,18 @@
 #include "src/execution/frames-inl.h"
 #include "src/execution/isolate-inl.h"
 #include "src/execution/isolate.h"
+#include "src/handles/handles.h"
 #include "src/heap/heap-inl.h"  // For ToBoolean. TODO(jkummerow): Drop.
+#include "src/interpreter/bytecode-flags-and-tokens.h"
 #include "src/objects/arguments-inl.h"
+#include "src/objects/fixed-array.h"
+#include "src/objects/heap-object.h"
+#include "src/objects/js-disposable-stack-inl.h"
+#include "src/objects/js-disposable-stack.h"
+#include "src/objects/objects.h"
+#include "src/objects/oddball.h"
 #include "src/objects/smi.h"
+#include "src/objects/tagged.h"
 #include "src/runtime/runtime-utils.h"
 
 namespace v8 {
@@ -25,12 +34,18 @@ RUNTIME_FUNCTION(Runtime_ThrowConstAssignError) {
                                  NewTypeError(MessageTemplate::kConstAssign));
 }
 
+RUNTIME_FUNCTION(Runtime_ThrowUsingAssignError) {
+  HandleScope scope(isolate);
+  THROW_NEW_ERROR_RETURN_FAILURE(isolate,
+                                 NewTypeError(MessageTemplate::kUsingAssign));
+}
+
 namespace {
 
 enum class RedeclarationType { kSyntaxError = 0, kTypeError = 1 };
 
-Object ThrowRedeclarationError(Isolate* isolate, Handle<String> name,
-                               RedeclarationType redeclaration_type) {
+Tagged<Object> ThrowRedeclarationError(Isolate* isolate, Handle<String> name,
+                                       RedeclarationType redeclaration_type) {
   HandleScope scope(isolate);
   if (redeclaration_type == RedeclarationType::kSyntaxError) {
     THROW_NEW_ERROR_RETURN_FAILURE(
@@ -42,11 +57,11 @@ Object ThrowRedeclarationError(Isolate* isolate, Handle<String> name,
 }
 
 // May throw a RedeclarationError.
-Object DeclareGlobal(Isolate* isolate, Handle<JSGlobalObject> global,
-                     Handle<String> name, Handle<Object> value,
-                     PropertyAttributes attr, bool is_var,
-                     RedeclarationType redeclaration_type) {
-  Handle<ScriptContextTable> script_contexts(
+Tagged<Object> DeclareGlobal(Isolate* isolate, Handle<JSGlobalObject> global,
+                             Handle<String> name, Handle<Object> value,
+                             PropertyAttributes attr, bool is_var,
+                             RedeclarationType redeclaration_type) {
+  DirectHandle<ScriptContextTable> script_contexts(
       global->native_context()->script_context_table(), isolate);
   VariableLookupResult lookup;
   if (script_contexts->Lookup(name, &lookup) &&
@@ -120,45 +135,41 @@ RUNTIME_FUNCTION(Runtime_DeclareModuleExports) {
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
 
-  Handle<FixedArray> declarations = args.at<FixedArray>(0);
-  Handle<JSFunction> closure = args.at<JSFunction>(1);
+  DirectHandle<FixedArray> declarations = args.at<FixedArray>(0);
+  DirectHandle<JSFunction> closure = args.at<JSFunction>(1);
 
-  Handle<ClosureFeedbackCellArray> closure_feedback_cell_array =
-      Handle<ClosureFeedbackCellArray>::null();
-  if (closure->has_feedback_vector()) {
-    closure_feedback_cell_array = Handle<ClosureFeedbackCellArray>(
-        closure->feedback_vector()->closure_feedback_cell_array(), isolate);
-  } else {
-    closure_feedback_cell_array = Handle<ClosureFeedbackCellArray>(
-        closure->closure_feedback_cell_array(), isolate);
-  }
+  DirectHandle<ClosureFeedbackCellArray> closure_feedback_cell_array(
+      closure->has_feedback_vector()
+          ? closure->feedback_vector()->closure_feedback_cell_array()
+          : closure->closure_feedback_cell_array(),
+      isolate);
 
   Handle<Context> context(isolate->context(), isolate);
   DCHECK(context->IsModuleContext());
-  Handle<FixedArray> exports(
-      SourceTextModule::cast(context->extension())->regular_exports(), isolate);
+  DirectHandle<FixedArray> exports(
+      Cast<SourceTextModule>(context->extension())->regular_exports(), isolate);
 
   int length = declarations->length();
   FOR_WITH_HANDLE_SCOPE(isolate, int, i = 0, i, i < length, i++, {
-    Object decl = declarations->get(i);
+    Tagged<Object> decl = declarations->get(i);
     int index;
-    Object value;
+    Tagged<Object> value;
     if (IsSmi(decl)) {
       index = Smi::ToInt(decl);
       value = ReadOnlyRoots(isolate).the_hole_value();
     } else {
       Handle<SharedFunctionInfo> sfi(
-          SharedFunctionInfo::cast(declarations->get(i)), isolate);
+          Cast<SharedFunctionInfo>(declarations->get(i)), isolate);
       int feedback_index = Smi::ToInt(declarations->get(++i));
       index = Smi::ToInt(declarations->get(++i));
-      Handle<FeedbackCell> feedback_cell =
-          closure_feedback_cell_array->GetFeedbackCell(feedback_index);
+      Handle<FeedbackCell> feedback_cell(
+          closure_feedback_cell_array->get(feedback_index), isolate);
       value = *Factory::JSFunctionBuilder(isolate, sfi, context)
                    .set_feedback_cell(feedback_cell)
                    .Build();
     }
 
-    Cell::cast(exports->get(index - 1))->set_value(value);
+    Cast<Cell>(exports->get(index - 1))->set_value(value);
   });
 
   return ReadOnlyRoots(isolate).undefined_value();
@@ -168,21 +179,17 @@ RUNTIME_FUNCTION(Runtime_DeclareGlobals) {
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
 
-  Handle<FixedArray> declarations = args.at<FixedArray>(0);
-  Handle<JSFunction> closure = args.at<JSFunction>(1);
+  DirectHandle<FixedArray> declarations = args.at<FixedArray>(0);
+  DirectHandle<JSFunction> closure = args.at<JSFunction>(1);
 
   Handle<JSGlobalObject> global(isolate->global_object());
   Handle<Context> context(isolate->context(), isolate);
 
-  Handle<ClosureFeedbackCellArray> closure_feedback_cell_array =
-      Handle<ClosureFeedbackCellArray>::null();
-  if (closure->has_feedback_vector()) {
-    closure_feedback_cell_array = Handle<ClosureFeedbackCellArray>(
-        closure->feedback_vector()->closure_feedback_cell_array(), isolate);
-  } else {
-    closure_feedback_cell_array = Handle<ClosureFeedbackCellArray>(
-        closure->closure_feedback_cell_array(), isolate);
-  }
+  DirectHandle<ClosureFeedbackCellArray> closure_feedback_cell_array(
+      closure->has_feedback_vector()
+          ? closure->feedback_vector()->closure_feedback_cell_array()
+          : closure->closure_feedback_cell_array(),
+      isolate);
 
   // Traverse the name/value pairs and set the properties.
   int length = declarations->length();
@@ -193,14 +200,14 @@ RUNTIME_FUNCTION(Runtime_DeclareGlobals) {
     bool is_var = IsString(*decl);
 
     if (is_var) {
-      name = Handle<String>::cast(decl);
+      name = Cast<String>(decl);
       value = isolate->factory()->undefined_value();
     } else {
-      Handle<SharedFunctionInfo> sfi = Handle<SharedFunctionInfo>::cast(decl);
+      Handle<SharedFunctionInfo> sfi = Cast<SharedFunctionInfo>(decl);
       name = handle(sfi->Name(), isolate);
       int index = Smi::ToInt(declarations->get(++i));
-      Handle<FeedbackCell> feedback_cell =
-          closure_feedback_cell_array->GetFeedbackCell(index);
+      Handle<FeedbackCell> feedback_cell(
+          closure_feedback_cell_array->get(index), isolate);
       value = Factory::JSFunctionBuilder(isolate, sfi, context)
                   .set_feedback_cell(feedback_cell)
                   .Build();
@@ -208,7 +215,7 @@ RUNTIME_FUNCTION(Runtime_DeclareGlobals) {
 
     // Compute the property attributes. According to ECMA-262,
     // the property must be non-configurable except in eval.
-    Script script = Script::cast(closure->shared()->script());
+    Tagged<Script> script = Cast<Script>(closure->shared()->script());
     PropertyAttributes attr =
         script->compilation_type() == Script::CompilationType::kEval
             ? NONE
@@ -216,18 +223,106 @@ RUNTIME_FUNCTION(Runtime_DeclareGlobals) {
 
     // ES#sec-globaldeclarationinstantiation 5.d:
     // If hasRestrictedGlobal is true, throw a SyntaxError exception.
-    Object result = DeclareGlobal(isolate, global, name, value, attr, is_var,
-                                  RedeclarationType::kSyntaxError);
-    if (isolate->has_pending_exception()) return result;
+    Tagged<Object> result =
+        DeclareGlobal(isolate, global, name, value, attr, is_var,
+                      RedeclarationType::kSyntaxError);
+    if (IsException(result)) return result;
   });
 
   return ReadOnlyRoots(isolate).undefined_value();
 }
 
+RUNTIME_FUNCTION(Runtime_InitializeDisposableStack) {
+  HandleScope scope(isolate);
+  DCHECK_EQ(0, args.length());
+
+  DirectHandle<JSDisposableStackBase> disposable_stack =
+      isolate->factory()->NewJSDisposableStackBase();
+  JSDisposableStackBase::InitializeJSDisposableStackBase(isolate,
+                                                         disposable_stack);
+  return *disposable_stack;
+}
+
+Tagged<Object> AddToDisposableStack(Isolate* isolate,
+                                    DirectHandle<JSDisposableStackBase> stack,
+                                    Handle<Object> value,
+                                    DisposeMethodCallType type,
+                                    DisposeMethodHint hint) {
+  // a. If V is either null or undefined and hint is sync-dispose, return
+  // unused.
+  if (IsNullOrUndefined(*value)) {
+    return *value;
+  }
+
+  Handle<Object> method;
+  ASSIGN_RETURN_FAILURE_ON_EXCEPTION(
+      isolate, method,
+      JSDisposableStackBase::CheckValueAndGetDisposeMethod(isolate, value,
+                                                           hint));
+
+  // Return the DisposableResource Record { [[ResourceValue]]: V, [[Hint]]:
+  // hint, [[DisposeMethod]]: method }.
+  JSDisposableStackBase::Add(isolate, stack, value, method, type, hint);
+  return *value;
+}
+
+RUNTIME_FUNCTION(Runtime_AddDisposableValue) {
+  HandleScope scope(isolate);
+  DCHECK_EQ(2, args.length());
+
+  DirectHandle<JSDisposableStackBase> stack = args.at<JSDisposableStackBase>(0);
+  Handle<Object> value = args.at<Object>(1);
+
+  // Return the DisposableResource Record { [[ResourceValue]]: V, [[Hint]]:
+  // hint, [[DisposeMethod]]: method }.
+  return AddToDisposableStack(isolate, stack, value,
+                              DisposeMethodCallType::kValueIsReceiver,
+                              DisposeMethodHint::kSyncDispose);
+}
+
+RUNTIME_FUNCTION(Runtime_AddAsyncDisposableValue) {
+  HandleScope scope(isolate);
+  DCHECK_EQ(2, args.length());
+
+  DirectHandle<JSDisposableStackBase> stack = args.at<JSDisposableStackBase>(0);
+  Handle<Object> value = args.at<Object>(1);
+
+  // Return the DisposableResource Record { [[ResourceValue]]: V, [[Hint]]:
+  // hint, [[DisposeMethod]]: method }.
+  return AddToDisposableStack(isolate, stack, value,
+                              DisposeMethodCallType::kValueIsReceiver,
+                              DisposeMethodHint::kAsyncDispose);
+}
+
+RUNTIME_FUNCTION(Runtime_DisposeDisposableStack) {
+  HandleScope scope(isolate);
+  DCHECK_EQ(4, args.length());
+
+  DirectHandle<JSDisposableStackBase> disposable_stack =
+      args.at<JSDisposableStackBase>(0);
+  DirectHandle<Smi> continuation_token = args.at<Smi>(1);
+  Handle<Object> continuation_error = args.at<Object>(2);
+  DirectHandle<Smi> has_await_using = args.at<Smi>(3);
+
+  Handle<Object> result;
+  ASSIGN_RETURN_FAILURE_ON_EXCEPTION(
+      isolate, result,
+      JSDisposableStackBase::DisposeResources(
+          isolate, disposable_stack,
+          (*continuation_token !=
+           Smi::FromInt(static_cast<int>(
+               interpreter::TryFinallyContinuationToken::kRethrowToken)))
+              ? MaybeHandle<Object>()
+              : continuation_error,
+          static_cast<DisposableStackResourcesType>(
+              Smi::ToInt(*has_await_using))));
+  return *result;
+}
+
 namespace {
 
-Object DeclareEvalHelper(Isolate* isolate, Handle<String> name,
-                         Handle<Object> value) {
+Tagged<Object> DeclareEvalHelper(Isolate* isolate, Handle<String> name,
+                                 Handle<Object> value) {
   // Declarations are always made in a function, native, eval, or script
   // context, or a declaration block scope. Since this is called from eval, the
   // context passed is the context of the caller, which may be some nested
@@ -259,25 +354,25 @@ Object DeclareEvalHelper(Isolate* isolate, Handle<String> name,
       Context::Lookup(context, name, DONT_FOLLOW_CHAINS, &index, &attributes,
                       &init_flag, &mode);
   DCHECK(holder.is_null() || !IsSourceTextModule(*holder));
-  DCHECK(!isolate->has_pending_exception());
+  DCHECK(!isolate->has_exception());
 
   Handle<JSObject> object;
 
   if (attributes != ABSENT && IsJSGlobalObject(*holder)) {
     // ES#sec-evaldeclarationinstantiation 8.a.iv.1.b:
     // If fnDefinable is false, throw a TypeError exception.
-    return DeclareGlobal(isolate, Handle<JSGlobalObject>::cast(holder), name,
-                         value, NONE, is_var, RedeclarationType::kTypeError);
+    return DeclareGlobal(isolate, Cast<JSGlobalObject>(holder), name, value,
+                         NONE, is_var, RedeclarationType::kTypeError);
   }
   if (context->has_extension() && IsJSGlobalObject(context->extension())) {
-    Handle<JSGlobalObject> global(JSGlobalObject::cast(context->extension()),
+    Handle<JSGlobalObject> global(Cast<JSGlobalObject>(context->extension()),
                                   isolate);
     return DeclareGlobal(isolate, global, name, value, NONE, is_var,
                          RedeclarationType::kTypeError);
   } else if (context->IsScriptContext()) {
     DCHECK(IsJSGlobalObject(context->global_object()));
     Handle<JSGlobalObject> global(
-        JSGlobalObject::cast(context->global_object()), isolate);
+        Cast<JSGlobalObject>(context->global_object()), isolate);
     return DeclareGlobal(isolate, global, name, value, NONE, is_var,
                          RedeclarationType::kTypeError);
   }
@@ -294,7 +389,7 @@ Object DeclareEvalHelper(Isolate* isolate, Handle<String> name,
       return ReadOnlyRoots(isolate).undefined_value();
     }
 
-    object = Handle<JSObject>::cast(holder);
+    object = Cast<JSObject>(holder);
 
   } else if (context->has_extension() && !is_debug_evaluate_in_module) {
     object = handle(context->extension_object(), isolate);
@@ -350,7 +445,7 @@ std::unique_ptr<Handle<Object>[]> GetCallerArguments(Isolate* isolate,
   // Find frame containing arguments passed to the caller.
   JavaScriptStackFrameIterator it(isolate);
   JavaScriptFrame* frame = it.frame();
-  std::vector<SharedFunctionInfo> functions;
+  std::vector<Tagged<SharedFunctionInfo>> functions;
   frame->GetFunctions(&functions);
   if (functions.size() > 1) {
     int inlined_jsframe_index = static_cast<int>(functions.size()) - 1;
@@ -418,11 +513,11 @@ Handle<JSObject> NewSloppyArguments(Isolate* isolate, Handle<JSFunction> callee,
 
       // Store the context and the arguments array at the beginning of the
       // parameter map.
-      Handle<Context> context(isolate->context(), isolate);
-      Handle<FixedArray> arguments = isolate->factory()->NewFixedArray(
+      DirectHandle<Context> context(isolate->context(), isolate);
+      DirectHandle<FixedArray> arguments = isolate->factory()->NewFixedArray(
           argument_count, AllocationType::kYoung);
 
-      Handle<SloppyArgumentsElements> parameter_map =
+      DirectHandle<SloppyArgumentsElements> parameter_map =
           isolate->factory()->NewSloppyArgumentsElements(
               mapped_count, context, arguments, AllocationType::kYoung);
 
@@ -438,7 +533,8 @@ Handle<JSObject> NewSloppyArguments(Isolate* isolate, Handle<JSFunction> callee,
         --index;
       }
 
-      Handle<ScopeInfo> scope_info(callee->shared()->scope_info(), isolate);
+      DirectHandle<ScopeInfo> scope_info(callee->shared()->scope_info(),
+                                         isolate);
 
       // First mark all mappable slots as unmapped and copy the values into the
       // arguments object.
@@ -450,18 +546,19 @@ Handle<JSObject> NewSloppyArguments(Isolate* isolate, Handle<JSFunction> callee,
 
       // Walk all context slots to find context allocated parameters. Mark each
       // found parameter as mapped.
+      ReadOnlyRoots roots{isolate};
       for (int i = 0; i < scope_info->ContextLocalCount(); i++) {
         if (!scope_info->ContextLocalIsParameter(i)) continue;
         int parameter = scope_info->ContextLocalParameterNumber(i);
         if (parameter >= mapped_count) continue;
-        arguments->set_the_hole(parameter);
-        Smi slot = Smi::FromInt(scope_info->ContextHeaderLength() + i);
+        arguments->set_the_hole(roots, parameter);
+        Tagged<Smi> slot = Smi::FromInt(scope_info->ContextHeaderLength() + i);
         parameter_map->set_mapped_entries(parameter, slot);
       }
     } else {
       // If there is no aliasing, the arguments object elements are not
       // special in any way.
-      Handle<FixedArray> elements = isolate->factory()->NewFixedArray(
+      DirectHandle<FixedArray> elements = isolate->factory()->NewFixedArray(
           argument_count, AllocationType::kYoung);
       result->set_elements(*elements);
       for (int i = 0; i < argument_count; ++i) {
@@ -475,7 +572,7 @@ Handle<JSObject> NewSloppyArguments(Isolate* isolate, Handle<JSFunction> callee,
 class HandleArguments {
  public:
   explicit HandleArguments(Handle<Object>* array) : array_(array) {}
-  Object operator[](int index) { return *array_[index]; }
+  Tagged<Object> operator[](int index) { return *array_[index]; }
 
  private:
   Handle<Object>* array_;
@@ -484,7 +581,7 @@ class HandleArguments {
 class ParameterArguments {
  public:
   explicit ParameterArguments(Address parameters) : parameters_(parameters) {}
-  Object operator[](int index) {
+  Tagged<Object> operator[](int index) {
     return *FullObjectSlot(parameters_ - (index + 1) * kSystemPointerSize);
   }
 
@@ -516,10 +613,10 @@ RUNTIME_FUNCTION(Runtime_NewStrictArguments) {
   int argument_count = 0;
   std::unique_ptr<Handle<Object>[]> arguments =
       GetCallerArguments(isolate, &argument_count);
-  Handle<JSObject> result =
+  DirectHandle<JSObject> result =
       isolate->factory()->NewArgumentsObject(callee, argument_count);
   if (argument_count) {
-    Handle<FixedArray> array =
+    DirectHandle<FixedArray> array =
         isolate->factory()->NewFixedArray(argument_count);
     DisallowGarbageCollection no_gc;
     WriteBarrierMode mode = array->GetWriteBarrierMode(no_gc);
@@ -535,7 +632,7 @@ RUNTIME_FUNCTION(Runtime_NewStrictArguments) {
 RUNTIME_FUNCTION(Runtime_NewRestParameter) {
   HandleScope scope(isolate);
   DCHECK_EQ(1, args.length());
-  Handle<JSFunction> callee = args.at<JSFunction>(0);
+  DirectHandle<JSFunction> callee = args.at<JSFunction>(0);
   int start_index =
       callee->shared()->internal_formal_parameter_count_without_receiver();
   // This generic runtime function can also be used when the caller has been
@@ -544,12 +641,13 @@ RUNTIME_FUNCTION(Runtime_NewRestParameter) {
   std::unique_ptr<Handle<Object>[]> arguments =
       GetCallerArguments(isolate, &argument_count);
   int num_elements = std::max(0, argument_count - start_index);
-  Handle<JSObject> result = isolate->factory()->NewJSArray(
+  DirectHandle<JSObject> result = isolate->factory()->NewJSArray(
       PACKED_ELEMENTS, num_elements, num_elements,
       ArrayStorageAllocationMode::DONT_INITIALIZE_ARRAY_ELEMENTS);
+  if (num_elements == 0) return *result;
   {
     DisallowGarbageCollection no_gc;
-    FixedArray elements = FixedArray::cast(result->elements());
+    Tagged<FixedArray> elements = Cast<FixedArray>(result->elements());
     WriteBarrierMode mode = elements->GetWriteBarrierMode(no_gc);
     for (int i = 0; i < num_elements; i++) {
       elements->set(i, *arguments[i + start_index], mode);
@@ -588,9 +686,9 @@ RUNTIME_FUNCTION(Runtime_NewFunctionContext) {
   HandleScope scope(isolate);
   DCHECK_EQ(1, args.length());
 
-  Handle<ScopeInfo> scope_info = args.at<ScopeInfo>(0);
+  DirectHandle<ScopeInfo> scope_info = args.at<ScopeInfo>(0);
 
-  Handle<Context> outer(isolate->context(), isolate);
+  DirectHandle<Context> outer(isolate->context(), isolate);
   return *isolate->factory()->NewFunctionContext(outer, scope_info);
 }
 
@@ -598,9 +696,9 @@ RUNTIME_FUNCTION(Runtime_NewFunctionContext) {
 RUNTIME_FUNCTION(Runtime_PushWithContext) {
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
-  Handle<JSReceiver> extension_object = args.at<JSReceiver>(0);
-  Handle<ScopeInfo> scope_info = args.at<ScopeInfo>(1);
-  Handle<Context> current(isolate->context(), isolate);
+  DirectHandle<JSReceiver> extension_object = args.at<JSReceiver>(0);
+  DirectHandle<ScopeInfo> scope_info = args.at<ScopeInfo>(1);
+  DirectHandle<Context> current(isolate->context(), isolate);
   return *isolate->factory()->NewWithContext(current, scope_info,
                                              extension_object);
 }
@@ -609,9 +707,9 @@ RUNTIME_FUNCTION(Runtime_PushWithContext) {
 RUNTIME_FUNCTION(Runtime_PushCatchContext) {
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
-  Handle<Object> thrown_object = args.at(0);
-  Handle<ScopeInfo> scope_info = args.at<ScopeInfo>(1);
-  Handle<Context> current(isolate->context(), isolate);
+  DirectHandle<Object> thrown_object = args.at(0);
+  DirectHandle<ScopeInfo> scope_info = args.at<ScopeInfo>(1);
+  DirectHandle<Context> current(isolate->context(), isolate);
   return *isolate->factory()->NewCatchContext(current, scope_info,
                                               thrown_object);
 }
@@ -620,8 +718,8 @@ RUNTIME_FUNCTION(Runtime_PushCatchContext) {
 RUNTIME_FUNCTION(Runtime_PushBlockContext) {
   HandleScope scope(isolate);
   DCHECK_EQ(1, args.length());
-  Handle<ScopeInfo> scope_info = args.at<ScopeInfo>(0);
-  Handle<Context> current(isolate->context(), isolate);
+  DirectHandle<ScopeInfo> scope_info = args.at<ScopeInfo>(0);
+  DirectHandle<Context> current(isolate->context(), isolate);
   return *isolate->factory()->NewBlockContext(current, scope_info);
 }
 
@@ -642,8 +740,7 @@ RUNTIME_FUNCTION(Runtime_DeleteLookupSlot) {
   // If the slot was not found the result is true.
   if (holder.is_null()) {
     // In case of JSProxy, an exception might have been thrown.
-    if (isolate->has_pending_exception())
-      return ReadOnlyRoots(isolate).exception();
+    if (isolate->has_exception()) return ReadOnlyRoots(isolate).exception();
     return ReadOnlyRoots(isolate).true_value();
   }
 
@@ -656,7 +753,7 @@ RUNTIME_FUNCTION(Runtime_DeleteLookupSlot) {
   // The slot was found in a JSReceiver, either a context extension object,
   // the global object, or the subject of a with.  Try to delete it
   // (respecting DONT_DELETE).
-  Handle<JSReceiver> object = Handle<JSReceiver>::cast(holder);
+  Handle<JSReceiver> object = Cast<JSReceiver>(holder);
   Maybe<bool> result = JSReceiver::DeleteProperty(object, name);
   MAYBE_RETURN(result, ReadOnlyRoots(isolate).exception());
   return isolate->heap()->ToBoolean(result.FromJust());
@@ -675,25 +772,24 @@ MaybeHandle<Object> LoadLookupSlot(Isolate* isolate, Handle<String> name,
   Handle<Context> context(isolate->context(), isolate);
   Handle<Object> holder = Context::Lookup(context, name, FOLLOW_CHAINS, &index,
                                           &attributes, &flag, &mode);
-  if (isolate->has_pending_exception()) return MaybeHandle<Object>();
+  if (isolate->has_exception()) return MaybeHandle<Object>();
 
   if (!holder.is_null() && IsSourceTextModule(*holder)) {
     Handle<Object> receiver = isolate->factory()->undefined_value();
     if (receiver_return) *receiver_return = receiver;
     return SourceTextModule::LoadVariable(
-        isolate, Handle<SourceTextModule>::cast(holder), index);
+        isolate, Cast<SourceTextModule>(holder), index);
   }
   if (index != Context::kNotFound) {
     DCHECK(IsContext(*holder));
     // If the "property" we were looking for is a local variable, the
     // receiver is the global object; see ECMA-262, 3rd., 10.1.6 and 10.2.3.
     Handle<Object> receiver = isolate->factory()->undefined_value();
-    Handle<Object> value = handle(Context::cast(*holder)->get(index), isolate);
+    Handle<Object> value = handle(Cast<Context>(*holder)->get(index), isolate);
     // Check for uninitialized bindings.
     if (flag == kNeedsInitialization && IsTheHole(*value, isolate)) {
       THROW_NEW_ERROR(isolate,
-                      NewReferenceError(MessageTemplate::kNotDefined, name),
-                      Object);
+                      NewReferenceError(MessageTemplate::kNotDefined, name));
     }
     DCHECK(!IsTheHole(*value, isolate));
     if (receiver_return) *receiver_return = receiver;
@@ -707,12 +803,12 @@ MaybeHandle<Object> LoadLookupSlot(Isolate* isolate, Handle<String> name,
     // No need to unhole the value here.  This is taken care of by the
     // GetProperty function.
     Handle<Object> value;
-    ASSIGN_RETURN_ON_EXCEPTION(
-        isolate, value, Object::GetProperty(isolate, holder, name), Object);
+    ASSIGN_RETURN_ON_EXCEPTION(isolate, value,
+                               Object::GetProperty(isolate, holder, name));
     if (receiver_return) {
       *receiver_return =
           (IsJSGlobalObject(*holder) || IsJSContextExtensionObject(*holder))
-              ? Handle<Object>::cast(isolate->factory()->undefined_value())
+              ? Cast<Object>(isolate->factory()->undefined_value())
               : holder;
     }
     return value;
@@ -720,8 +816,8 @@ MaybeHandle<Object> LoadLookupSlot(Isolate* isolate, Handle<String> name,
 
   if (should_throw == kThrowOnError) {
     // The property doesn't exist - throw exception.
-    THROW_NEW_ERROR(
-        isolate, NewReferenceError(MessageTemplate::kNotDefined, name), Object);
+    THROW_NEW_ERROR(isolate,
+                    NewReferenceError(MessageTemplate::kNotDefined, name));
   }
 
   // The property doesn't exist - return undefined.
@@ -757,7 +853,7 @@ RUNTIME_FUNCTION_RETURN_PAIR(Runtime_LoadLookupSlotForCall) {
   Handle<Object> receiver;
   ASSIGN_RETURN_ON_EXCEPTION_VALUE(
       isolate, value, LoadLookupSlot(isolate, name, kThrowOnError, &receiver),
-      MakePair(ReadOnlyRoots(isolate).exception(), Object()));
+      MakePair(ReadOnlyRoots(isolate).exception(), Tagged<Object>()));
   return MakePair(*value, *receiver);
 }
 
@@ -772,9 +868,9 @@ RUNTIME_FUNCTION(Runtime_LoadLookupSlotForCall_Baseline) {
   Handle<Object> value;
   if (!LoadLookupSlot(isolate, name, kThrowOnError, &receiver)
            .ToHandle(&value)) {
-    DCHECK((isolate)->has_pending_exception());
+    DCHECK((isolate)->has_exception());
     value_ret.store(ReadOnlyRoots(isolate).exception());
-    receiver_ret.store(Object());
+    receiver_ret.store(Tagged<Object>());
     return ReadOnlyRoots(isolate).exception();
   }
   value_ret.store(*value);
@@ -798,30 +894,34 @@ MaybeHandle<Object> StoreLookupSlot(
                       &flag, &mode, &is_sloppy_function_name);
   if (holder.is_null()) {
     // In case of JSProxy, an exception might have been thrown.
-    if (isolate->has_pending_exception()) return MaybeHandle<Object>();
+    if (isolate->has_exception()) return MaybeHandle<Object>();
   } else if (IsSourceTextModule(*holder)) {
     if ((attributes & READ_ONLY) == 0) {
-      SourceTextModule::StoreVariable(Handle<SourceTextModule>::cast(holder),
-                                      index, value);
+      SourceTextModule::StoreVariable(Cast<SourceTextModule>(holder), index,
+                                      value);
     } else {
-      THROW_NEW_ERROR(
-          isolate, NewTypeError(MessageTemplate::kConstAssign, name), Object);
+      THROW_NEW_ERROR(isolate,
+                      NewTypeError(MessageTemplate::kConstAssign, name));
     }
     return value;
   }
   // The property was found in a context slot.
   if (index != Context::kNotFound) {
+    auto holder_context = Cast<Context>(holder);
     if (flag == kNeedsInitialization &&
-        IsTheHole(Handle<Context>::cast(holder)->get(index), isolate)) {
+        IsTheHole(holder_context->get(index), isolate)) {
       THROW_NEW_ERROR(isolate,
-                      NewReferenceError(MessageTemplate::kNotDefined, name),
-                      Object);
+                      NewReferenceError(MessageTemplate::kNotDefined, name));
     }
     if ((attributes & READ_ONLY) == 0) {
-      Handle<Context>::cast(holder)->set(index, *value);
+      if (v8_flags.const_tracking_let && holder_context->IsScriptContext()) {
+        Context::UpdateConstTrackingLetSideData(holder_context, index, value,
+                                                isolate);
+      }
+      Cast<Context>(holder)->set(index, *value);
     } else if (!is_sloppy_function_name || is_strict(language_mode)) {
-      THROW_NEW_ERROR(
-          isolate, NewTypeError(MessageTemplate::kConstAssign, name), Object);
+      THROW_NEW_ERROR(isolate,
+                      NewTypeError(MessageTemplate::kConstAssign, name));
     }
     return value;
   }
@@ -832,19 +932,18 @@ MaybeHandle<Object> StoreLookupSlot(
   Handle<JSReceiver> object;
   if (attributes != ABSENT) {
     // The property exists on the holder.
-    object = Handle<JSReceiver>::cast(holder);
+    object = Cast<JSReceiver>(holder);
   } else if (is_strict(language_mode)) {
     // If absent in strict mode: throw.
-    THROW_NEW_ERROR(
-        isolate, NewReferenceError(MessageTemplate::kNotDefined, name), Object);
+    THROW_NEW_ERROR(isolate,
+                    NewReferenceError(MessageTemplate::kNotDefined, name));
   } else {
     // If absent in sloppy mode: add the property to the global object.
     object = handle(context->global_object(), isolate);
   }
 
   ASSIGN_RETURN_ON_EXCEPTION(isolate, value,
-                             Object::SetProperty(isolate, object, name, value),
-                             Object);
+                             Object::SetProperty(isolate, object, name, value));
   return value;
 }
 
@@ -893,17 +992,25 @@ RUNTIME_FUNCTION(Runtime_StoreGlobalNoHoleCheckForReplLetOrConst) {
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
   Handle<String> name = args.at<String>(0);
-  Handle<Object> value = args.at(1);
+  DirectHandle<Object> value = args.at(1);
 
-  Handle<Context> native_context = isolate->native_context();
-  Handle<ScriptContextTable> script_contexts(
+  DirectHandle<Context> native_context = isolate->native_context();
+  DirectHandle<ScriptContextTable> script_contexts(
       native_context->script_context_table(), isolate);
 
   VariableLookupResult lookup_result;
   bool found = script_contexts->Lookup(name, &lookup_result);
   CHECK(found);
-  Handle<Context> script_context = ScriptContextTable::GetContext(
-      isolate, script_contexts, lookup_result.context_index);
+  DirectHandle<Context> script_context(
+      script_contexts->get(lookup_result.context_index), isolate);
+  // We need to initialize the side data also for variables declared with
+  // VariableMode::kConst. This is because such variables can be accessed
+  // by functions using the LdaContextSlot bytecode, and such accesses are not
+  // regarded as "immutable" when optimizing.
+  if (v8_flags.const_tracking_let) {
+    Context::UpdateConstTrackingLetSideData(
+        script_context, lookup_result.slot_index, value, isolate);
+  }
 
   script_context->set(lookup_result.slot_index, *value);
   return *value;
