@@ -908,6 +908,10 @@ OutgoingMessage.prototype.write = function write(chunk, encoding, callback) {
 };
 
 function onError(msg, err, callback) {
+  if (msg.destroyed) {
+    return;
+  }
+
   const triggerAsyncId = msg.socket ? msg.socket[async_id_symbol] : undefined;
   defaultTriggerAsyncIdScope(triggerAsyncId,
                              process.nextTick,
@@ -919,7 +923,7 @@ function onError(msg, err, callback) {
 
 function emitErrorNt(msg, err, callback) {
   callback(err);
-  if (typeof msg.emit === 'function' && !msg._closed) {
+  if (typeof msg.emit === 'function' && !msg.destroyed) {
     msg.emit('error', err);
   }
 }
