@@ -97,6 +97,7 @@ let debug = require('internal/util/debuglog').debuglog('http', (fn) => {
 
 const dc = require('diagnostics_channel');
 const onRequestStartChannel = dc.channel('http.server.request.start');
+const onResponseCreatedChannel = dc.channel('http.server.response.created');
 const onResponseFinishChannel = dc.channel('http.server.response.finish');
 
 const kServerResponse = Symbol('ServerResponse');
@@ -224,6 +225,12 @@ function ServerResponse(req, options) {
     this._traceEventId = getNextTraceEventId();
     traceBegin(HTTP_SERVER_TRACE_EVENT_NAME, this._traceEventId);
   }
+  if (onResponseCreatedChannel.hasSubscribers) {
+    onResponseCreatedChannel.publish({
+      request: req,
+      response: this,
+    });
+  }
 }
 ObjectSetPrototypeOf(ServerResponse.prototype, OutgoingMessage.prototype);
 ObjectSetPrototypeOf(ServerResponse, OutgoingMessage);
