@@ -37,20 +37,20 @@ const {
   ObjectAssign,
   ObjectDefineProperty,
   ObjectPrototypeHasOwnProperty,
+  PromiseWithResolvers,
   RegExpPrototypeExec,
   SafeSet,
   StringPrototypeIncludes,
   StringPrototypeSlice,
   StringPrototypeToUpperCase,
+  SymbolDispose,
 } = primordials;
 
 const {
   convertToValidSignal,
-  createDeferredPromise,
   getSystemErrorName,
   kEmptyObject,
   promisify,
-  SymbolDispose,
 } = require('internal/util');
 const { isArrayBufferView } = require('internal/util/types');
 let debug = require('internal/util/debuglog').debuglog(
@@ -138,7 +138,7 @@ function fork(modulePath, args = [], options) {
     validateObject(options, 'options');
   }
   options = { __proto__: null, ...options, shell: false };
-  options.execPath = options.execPath || process.execPath;
+  options.execPath ||= process.execPath;
   validateArgumentNullCheck(options.execPath, 'options.execPath');
 
   // Prepare arguments for fork:
@@ -237,7 +237,7 @@ function exec(command, options, callback) {
 
 const customPromiseExecFunction = (orig) => {
   return (...args) => {
-    const { promise, resolve, reject } = createDeferredPromise();
+    const { promise, resolve, reject } = PromiseWithResolvers();
 
     promise.child = orig(...args, (err, stdout, stderr) => {
       if (err !== null) {
@@ -272,9 +272,7 @@ function normalizeExecFileArgs(file, args, options, callback) {
     args = null;
   }
 
-  if (args == null) {
-    args = [];
-  }
+  args ??= [];
 
   if (typeof options === 'function') {
     callback = options;
@@ -282,9 +280,7 @@ function normalizeExecFileArgs(file, args, options, callback) {
     validateObject(options, 'options');
   }
 
-  if (options == null) {
-    options = kEmptyObject;
-  }
+  options ??= kEmptyObject;
 
   if (callback != null) {
     validateFunction(callback, 'callback');
@@ -392,8 +388,7 @@ function execFile(file, args, options, callback) {
     let stderr;
     if (encoding ||
       (
-        child.stdout &&
-        child.stdout.readableEncoding
+        child.stdout?.readableEncoding
       )) {
       stdout = ArrayPrototypeJoin(_stdout, '');
     } else {
@@ -401,8 +396,7 @@ function execFile(file, args, options, callback) {
     }
     if (encoding ||
       (
-        child.stderr &&
-        child.stderr.readableEncoding
+        child.stderr?.readableEncoding
       )) {
       stderr = ArrayPrototypeJoin(_stderr, '');
     } else {
@@ -417,13 +411,11 @@ function execFile(file, args, options, callback) {
     if (args?.length)
       cmd += ` ${ArrayPrototypeJoin(args, ' ')}`;
 
-    if (!ex) {
-      ex = genericNodeError(`Command failed: ${cmd}\n${stderr}`, {
-        code: code < 0 ? getSystemErrorName(code) : code,
-        killed: child.killed || killed,
-        signal: signal,
-      });
-    }
+    ex ||= genericNodeError(`Command failed: ${cmd}\n${stderr}`, {
+      code: code < 0 ? getSystemErrorName(code) : code,
+      killed: child.killed || killed,
+      signal: signal,
+    });
 
     ex.cmd = cmd;
     callback(ex, stdout, stderr);
@@ -855,7 +847,7 @@ function spawnSync(file, args, options) {
 
   // We may want to pass data in on any given fd, ensure it is a valid buffer
   for (let i = 0; i < options.stdio.length; i++) {
-    const input = options.stdio[i] && options.stdio[i].input;
+    const input = options.stdio[i]?.input;
     if (input != null) {
       const pipe = options.stdio[i] = { ...options.stdio[i] };
       if (isArrayBufferView(input)) {
