@@ -276,6 +276,12 @@ ObjectDefineProperty(EventEmitter, 'defaultMaxListeners', {
   },
 });
 
+function hasEventListener(self, type) {
+  if (type === undefined)
+    return self._events !== undefined;
+  return self._events !== undefined && self._events[type] !== undefined;
+};
+
 ObjectDefineProperties(EventEmitter, {
   kMaxEventTargetListeners: {
     __proto__: null,
@@ -669,13 +675,11 @@ EventEmitter.prototype.removeListener =
     function removeListener(type, listener) {
       checkListener(listener);
 
-      const events = this._events;
-      if (events === undefined)
+      if (!hasEventListener(this, type))
         return this;
 
+      const events = this._events;
       const list = events[type];
-      if (list === undefined)
-        return this;
 
       if (list === listener || list.listener === listener) {
         this._eventsCount -= 1;
@@ -729,9 +733,9 @@ EventEmitter.prototype.off = EventEmitter.prototype.removeListener;
  */
 EventEmitter.prototype.removeAllListeners =
     function removeAllListeners(type) {
-      const events = this._events;
-      if (events === undefined)
+      if (!hasEventListener(this))
         return this;
+      const events = this._events;
 
       // Not listening for removeListener, no need to emit
       if (events.removeListener === undefined) {
@@ -776,14 +780,10 @@ EventEmitter.prototype.removeAllListeners =
     };
 
 function _listeners(target, type, unwrap) {
-  const events = target._events;
-
-  if (events === undefined)
+  if (!hasEventListener(target, type))
     return [];
 
-  const evlistener = events[type];
-  if (evlistener === undefined)
-    return [];
+  const evlistener = target._events[type];
 
   if (typeof evlistener === 'function')
     return unwrap ? [evlistener.listener || evlistener] : [evlistener];
@@ -958,10 +958,10 @@ function getMaxListeners(emitterOrTarget) {
  */
 async function once(emitter, name, options = kEmptyObject) {
   validateObject(options, 'options');
-  const signal = options?.signal;
+  const { signal } = options;
   validateAbortSignal(signal, 'options.signal');
   if (signal?.aborted)
-    throw new AbortError(undefined, { cause: signal?.reason });
+    throw new AbortError(undefined, { cause: signal.reason });
   return new Promise((resolve, reject) => {
     const errorListener = (err) => {
       emitter.removeListener(name, resolver);
@@ -1049,7 +1049,7 @@ function on(emitter, event, options = kEmptyObject) {
   const signal = options.signal;
   validateAbortSignal(signal, 'options.signal');
   if (signal?.aborted)
-    throw new AbortError(undefined, { cause: signal?.reason });
+    throw new AbortError(undefined, { cause: signal.reason });
   // Support both highWaterMark and highWatermark for backward compatibility
   const highWatermark = options.highWaterMark ?? options.highWatermark ?? NumberMAX_SAFE_INTEGER;
   validateInteger(highWatermark, 'options.highWaterMark', 1);
