@@ -17,7 +17,6 @@ const {
   ArrayPrototypeSort,
   ObjectAssign,
   PromisePrototypeThen,
-  PromiseResolve,
   PromiseWithResolvers,
   SafeMap,
   SafePromiseAll,
@@ -801,9 +800,17 @@ function run(options = kEmptyObject) {
     }
   }
 
-  const setupPromise = PromiseResolve(setup?.(root.reporter));
-  PromisePrototypeThen(PromisePrototypeThen(PromisePrototypeThen(setupPromise, runFiles), postRun), teardown);
+  const runChain = async () => {
+    if (typeof setup === 'function') {
+      await setup(root.reporter);
+    }
 
+    await runFiles();
+    postRun?.();
+    teardown?.();
+  };
+
+  runChain();
   return root.reporter;
 }
 
