@@ -16,7 +16,6 @@ const {
   ObjectSetPrototypeOf,
   ReflectGetOwnPropertyDescriptor,
   ReflectOwnKeys,
-  RegExpPrototypeSymbolReplace,
   SafeMap,
   SafeSet,
   StringPrototypeCharAt,
@@ -779,6 +778,8 @@ function isURL(self) {
  * for invalid URL inputs.
  */
 const kParseURLSymbol = Symbol('kParseURL');
+const kCreateURLFromPosixPathSymbol = Symbol('kCreateURLFromPosixPath');
+const kCreateURLFromWindowsPathSymbol = Symbol('kCreateURLFromWindowsPath');
 
 class URL {
   #context = new URLContext();
@@ -812,8 +813,17 @@ class URL {
       base = `${base}`;
     }
 
-    const raiseException = parseSymbol !== kParseURLSymbol;
-    const href = bindingUrl.parse(input, base, raiseException);
+    let href;
+    if (arguments.length < 3) {
+      href = bindingUrl.parse(input, base, true);
+    } else {
+      const raiseException = parseSymbol !== kParseURLSymbol;
+      const interpretAsWindowsPath = parseSymbol === kCreateURLFromWindowsPathSymbol;
+      const pathToFileURL = interpretAsWindowsPath || (parseSymbol === kCreateURLFromPosixPathSymbol);
+      href = pathToFileURL ?
+        bindingUrl.pathToFileURL(input, interpretAsWindowsPath, base) :
+        bindingUrl.parse(input, base, raiseException);
+    }
     if (href) {
       this.#updateContext(href);
     }
@@ -1500,76 +1510,9 @@ function fileURLToPath(path, options = kEmptyObject) {
   return (windows ?? isWindows) ? getPathFromURLWin32(path) : getPathFromURLPosix(path);
 }
 
-// RFC1738 defines the following chars as "unsafe" for URLs
-// @see https://www.ietf.org/rfc/rfc1738.txt 2.2. URL Character Encoding Issues
-const percentRegEx = /%/g;
-const newlineRegEx = /\n/g;
-const carriageReturnRegEx = /\r/g;
-const tabRegEx = /\t/g;
-const quoteRegEx = /"/g;
-const hashRegex = /#/g;
-const spaceRegEx = / /g;
-const questionMarkRegex = /\?/g;
-const openSquareBracketRegEx = /\[/g;
-const backslashRegEx = /\\/g;
-const closeSquareBracketRegEx = /]/g;
-const caretRegEx = /\^/g;
-const verticalBarRegEx = /\|/g;
-const tildeRegEx = /~/g;
-
-function encodePathChars(filepath, options = kEmptyObject) {
-  if (StringPrototypeIncludes(filepath, '%')) {
-    filepath = RegExpPrototypeSymbolReplace(percentRegEx, filepath, '%25');
-  }
-
-  if (StringPrototypeIncludes(filepath, '\t')) {
-    filepath = RegExpPrototypeSymbolReplace(tabRegEx, filepath, '%09');
-  }
-  if (StringPrototypeIncludes(filepath, '\n')) {
-    filepath = RegExpPrototypeSymbolReplace(newlineRegEx, filepath, '%0A');
-  }
-  if (StringPrototypeIncludes(filepath, '\r')) {
-    filepath = RegExpPrototypeSymbolReplace(carriageReturnRegEx, filepath, '%0D');
-  }
-  if (StringPrototypeIncludes(filepath, ' ')) {
-    filepath = RegExpPrototypeSymbolReplace(spaceRegEx, filepath, '%20');
-  }
-  if (StringPrototypeIncludes(filepath, '"')) {
-    filepath = RegExpPrototypeSymbolReplace(quoteRegEx, filepath, '%22');
-  }
-  if (StringPrototypeIncludes(filepath, '#')) {
-    filepath = RegExpPrototypeSymbolReplace(hashRegex, filepath, '%23');
-  }
-  if (StringPrototypeIncludes(filepath, '?')) {
-    filepath = RegExpPrototypeSymbolReplace(questionMarkRegex, filepath, '%3F');
-  }
-  if (StringPrototypeIncludes(filepath, '[')) {
-    filepath = RegExpPrototypeSymbolReplace(openSquareBracketRegEx, filepath, '%5B');
-  }
-  // Back-slashes must be special-cased on Windows, where they are treated as path separator.
-  if (!options.windows && StringPrototypeIncludes(filepath, '\\')) {
-    filepath = RegExpPrototypeSymbolReplace(backslashRegEx, filepath, '%5C');
-  }
-  if (StringPrototypeIncludes(filepath, ']')) {
-    filepath = RegExpPrototypeSymbolReplace(closeSquareBracketRegEx, filepath, '%5D');
-  }
-  if (StringPrototypeIncludes(filepath, '^')) {
-    filepath = RegExpPrototypeSymbolReplace(caretRegEx, filepath, '%5E');
-  }
-  if (StringPrototypeIncludes(filepath, '|')) {
-    filepath = RegExpPrototypeSymbolReplace(verticalBarRegEx, filepath, '%7C');
-  }
-  if (StringPrototypeIncludes(filepath, '~')) {
-    filepath = RegExpPrototypeSymbolReplace(tildeRegEx, filepath, '%7E');
-  }
-
-  return filepath;
-}
-
 function pathToFileURL(filepath, options = kEmptyObject) {
   const windows = options?.windows ?? isWindows;
   if (windows && StringPrototypeStartsWith(filepath, '\\\\')) {
-    const outURL = new URL('file://');
     // UNC path format: \\server\share\resource
     // Handle extended UNC path and standard UNC path
     // "\\?\UNC\" path prefix should be ignored.
@@ -1592,12 +1535,7 @@ function pathToFileURL(filepath, options = kEmptyObject) {
       );
     }
     const hostname = StringPrototypeSlice(filepath, prefixLength, hostnameEndIndex);
-    outURL.hostname = domainToASCII(hostname);
-    outURL.pathname = encodePathChars(
-      RegExpPrototypeSymbolReplace(backslashRegEx, StringPrototypeSlice(filepath, hostnameEndIndex), '/'),
-      { windows },
-    );
-    return outURL;
+    return new URL(StringPrototypeSlice(filepath, hostnameEndIndex), hostname, kCreateURLFromWindowsPathSymbol);
   }
   let resolved = windows ? path.win32.resolve(filepath) : path.posix.resolve(filepath);
   // path.resolve strips trailing slashes so we must add them back
@@ -1608,7 +1546,7 @@ function pathToFileURL(filepath, options = kEmptyObject) {
       resolved[resolved.length - 1] !== path.sep)
     resolved += '/';
 
-  return new URL(`file://${encodePathChars(resolved, { windows })}`);
+  return new URL(resolved, undefined, windows ? kCreateURLFromWindowsPathSymbol : kCreateURLFromPosixPathSymbol);
 }
 
 function toPathIfFileURL(fileURLOrPath) {
