@@ -11,6 +11,7 @@ const {
   PromisePrototypeThen,
   PromiseReject,
   PromiseResolve,
+  PromiseWithResolvers,
   Symbol,
   SymbolToStringTag,
 } = primordials;
@@ -30,7 +31,6 @@ const {
 } = internalBinding('messaging');
 
 const {
-  createDeferredPromise,
   customInspectSymbol: kInspect,
   kEmptyObject,
   kEnumerableProperty,
@@ -170,7 +170,7 @@ class WritableStream {
 
     this[kState] = createWritableStreamState();
 
-    this[kIsClosedPromise] = createDeferredPromise();
+    this[kIsClosedPromise] = PromiseWithResolvers();
     this[kControllerErrorFunction] = () => {};
 
     const size = extractSizeAlgorithm(strategy?.size);
@@ -300,11 +300,12 @@ ObjectDefineProperties(WritableStream.prototype, {
 });
 
 function InternalTransferredWritableStream() {
+  ObjectSetPrototypeOf(this, WritableStream.prototype);
   markTransferMode(this, false, true);
   this[kType] = 'WritableStream';
   this[kState] = createWritableStreamState();
 
-  this[kIsClosedPromise] = createDeferredPromise();
+  this[kIsClosedPromise] = PromiseWithResolvers();
 }
 
 ObjectSetPrototypeOf(InternalTransferredWritableStream.prototype, WritableStream.prototype);
@@ -516,10 +517,11 @@ ObjectDefineProperties(WritableStreamDefaultController.prototype, {
 });
 
 function InternalWritableStream(start, write, close, abort, highWaterMark, size) {
+  ObjectSetPrototypeOf(this, WritableStream.prototype);
   markTransferMode(this, false, true);
   this[kType] = 'WritableStream';
   this[kState] = createWritableStreamState();
-  this[kIsClosedPromise] = createDeferredPromise();
+  this[kIsClosedPromise] = PromiseWithResolvers();
 
   const controller = new WritableStreamDefaultController(kSkipThrow);
   setupWritableStreamDefaultController(
@@ -555,7 +557,7 @@ const isWritableStreamDefaultController =
 function createWritableStreamState() {
   return {
     __proto__: null,
-    close: createDeferredPromise(),
+    close: PromiseWithResolvers(),
     closeRequest: {
       __proto__: null,
       promise: undefined,
@@ -614,7 +616,7 @@ function setupWritableStreamDefaultWriter(writer, stream) {
     case 'writable':
       if (!writableStreamCloseQueuedOrInFlight(stream) &&
           stream[kState].backpressure) {
-        writer[kState].ready = createDeferredPromise();
+        writer[kState].ready = PromiseWithResolvers();
       } else {
         writer[kState].ready = {
           promise: PromiseResolve(),
@@ -661,7 +663,7 @@ function setupWritableStreamDefaultWriter(writer, stream) {
   }
 
   function setClosedPromiseToNewPromise() {
-    writer[kState].close = createDeferredPromise();
+    writer[kState].close = PromiseWithResolvers();
   }
 }
 
@@ -686,7 +688,7 @@ function writableStreamAbort(stream, reason) {
     reason = undefined;
   }
 
-  const abort = createDeferredPromise();
+  const abort = PromiseWithResolvers();
 
   stream[kState].pendingAbortRequest = {
     abort,
@@ -713,7 +715,7 @@ function writableStreamClose(stream) {
   }
   assert(state === 'writable' || state === 'erroring');
   assert(!writableStreamCloseQueuedOrInFlight(stream));
-  stream[kState].closeRequest = createDeferredPromise();
+  stream[kState].closeRequest = PromiseWithResolvers();
   const { promise } = stream[kState].closeRequest;
   if (writer !== undefined && backpressure && state === 'writable')
     writer[kState].ready.resolve?.();
@@ -729,7 +731,7 @@ function writableStreamUpdateBackpressure(stream, backpressure) {
   } = stream[kState];
   if (writer !== undefined && stream[kState].backpressure !== backpressure) {
     if (backpressure) {
-      writer[kState].ready = createDeferredPromise();
+      writer[kState].ready = PromiseWithResolvers();
     } else {
       writer[kState].ready.resolve?.();
     }
@@ -959,7 +961,7 @@ function writableStreamAddWriteRequest(stream) {
     promise,
     resolve,
     reject,
-  } = createDeferredPromise();
+  } = PromiseWithResolvers();
   ArrayPrototypePush(
     stream[kState].writeRequests,
     {
